#include "proxyserver.h"
#include "cmdline.h"

static struct gengetopt_args_info ai;

static uv_loop_t *loop;

static uv_timer_t reconnect_connection_timer;

static uv_tcp_t server;

static http_parser_settings settings;

static http_parser_settings api_settings;

evt_ctx_t ssl_ctx;

uv_mutex_t ssl_ctx_mutex;

uv_mutex_t loop_mutex;

uv_mutex_t connections_mutex;

uv_mutex_t messages_mutex;

LIST_HEAD(connected_list, api_connection_s) connected_head = LIST_HEAD_INITIALIZER(api_connection_s);

LIST_HEAD(unconnected_list, api_connection_s) unconnected_head = LIST_HEAD_INITIALIZER(api_connection_s);

static api_connection_t *api_connections;

LIST_HEAD(free_messages_list, message_s) free_messages_head = LIST_HEAD_INITIALIZER(message_s);

TAILQ_HEAD(used_messages_list, message_s) used_messages_head = TAILQ_HEAD_INITIALIZER(used_messages_head);

static message_t *messages;

uv_async_t send_message;

static int working = 1;

uv_signal_t signal_term;
uv_signal_t signal_kill;

void signal_handler_cb(uv_signal_t* handle, int signum) {
    fprintf(stderr, "Handle signal %d, stoping\n", signum);
    working = 0;
    uv_stop(loop);
}

int lookup_addr(char *ip, int port, struct sockaddr_in *addr) {
    if (UV_EINVAL != uv_ip4_addr(ip, port, addr)) {
        return 0;
    }

    struct hostent *he = gethostbyname2(ip, AF_INET);
    if (!he) {
        fprintf(stderr, "gethostbyname2(%s, AF_INET): %s\n", ip, strerror(errno));
        return -1;
    }

    struct in_addr **addr_list = (struct in_addr **) he->h_addr_list;
    int i;
    for (i = 0; addr_list[i] != NULL; i++) {
        if (UV_EINVAL != uv_ip4_addr(inet_ntoa(*addr_list[i]), port, addr)) {
            return 0;
        }
    }

    fprintf(stderr, "%s hostname doesn't have any IPv4 address\n", ip);
    return -1;
}

int main(int argc, char **argv) {
    int i;
    if (cmdline_parser(argc, argv, &ai) != 0) {
        exit(1);
    }

    struct sockaddr_in local_addr;
    if (lookup_addr(ai.local_address_arg, ai.local_port_arg, &local_addr) != 0) {
        exit(2);
    }

    if (evt_ctx_init(&ssl_ctx) < 0) {
        fprintf(stderr, "evt_ctx_init error\n");
        exit(3);
    }

    evt_ctx_set_nio(&ssl_ctx, NULL, uv_tls_writer);

    srand(time(NULL));

    LIST_INIT(&connected_head);
    LIST_INIT(&unconnected_head);

    uv_mutex_init(&ssl_ctx_mutex);

    uv_mutex_init(&connections_mutex);

    uv_mutex_init(&messages_mutex);

    uv_mutex_init(&loop_mutex);

    loop = uv_default_loop();

    uv_signal_init(loop, &signal_term);
    uv_signal_start(&signal_term, signal_handler_cb, SIGTERM);

    uv_signal_init(loop, &signal_kill);
    uv_signal_start(&signal_kill, signal_handler_cb, SIGKILL);

    if (ai.max_remote_connections_arg > 0) {
        api_connections = calloc(ai.max_remote_connections_arg, sizeof(api_connection_t));
        if (api_connections == NULL) {
            fprintf(stderr, "calloc(%d, sizeof(api_connection_t)) failed\n", ai.max_remote_connections_arg);
            exit(3);
        }

        for (i = 0; i < ai.max_remote_connections_arg; i++) {
            api_connection_open(&api_connections[i]);
        }
    }

    if (ai.max_messages_queue_length_arg < ai.threads_arg + 1) {
        fprintf(stderr, "messages queue should be > %d\n", ai.threads_arg + 1);
        exit(3);
    }

    messages = calloc(ai.max_messages_queue_length_arg, sizeof(message_t));
    if (messages == NULL) {
        fprintf(stderr, "calloc(%d, sizeof(messages_t)) failed\n", ai.max_messages_queue_length_arg);
        exit(3);
    }

    for (i = 0; i < ai.max_messages_queue_length_arg; i++) {
        LIST_INSERT_HEAD(&free_messages_head, &messages[i], free_messages);
    }

    uv_async_init(loop, &send_message, send_message_cb);

    http_parser_settings_init(&settings);
    http_parser_settings_init(&api_settings);

    settings.on_url = http_url_cb;
    settings.on_body = http_body_cb;
    settings.on_header_field = http_header_field_cb;
    settings.on_header_value = http_header_value_cb;
    settings.on_headers_complete = http_headers_complete_cb;
    settings.on_message_begin = http_message_begin_cb;

    api_settings.on_body = api_http_body_cb;
    api_settings.on_header_field = api_http_header_field_cb;
    api_settings.on_header_value = api_http_header_value_cb;
    api_settings.on_headers_complete = api_http_headers_complete_cb;
    api_settings.on_message_begin = api_http_message_begin_cb;

    uv_tcp_init(loop, &server);
    uv_tcp_bind(&server, (const struct sockaddr *) &local_addr, 0);

    int r = uv_listen((uv_stream_t *) &server, 128, tcp_new_connection_cb);

    if (r < 0) {
        fprintf(stderr, "Error on tcp listen: %s.\n", uv_strerror(r));
        exit(5);
    }

    if (ai.remote_reconnect_timeout_arg > 0) {
        uv_timer_init(loop, &reconnect_connection_timer);
        uv_timer_start(&reconnect_connection_timer, reconnect_connection_timer_cb,
                       ai.remote_reconnect_timeout_arg, ai.remote_reconnect_timeout_arg);
    }

    fprintf(stderr, "Listen to: %s:%d\n", ai.local_address_arg, ai.local_port_arg);

    if (ai.threads_arg > 0) {
        uv_thread_t *threads = calloc(ai.threads_arg, sizeof(uv_thread_t));
        if (threads == NULL) {
            fprintf(stderr, "evt_ctx_init error\n");
            exit(6);
        }
        for (i = 0; i < ai.threads_arg; i++) {
            uv_thread_create(&threads[i], thread_worker, loop);
        }

        for (i = 0; i < ai.threads_arg; i++) {
            uv_thread_join(&threads[i]);
        }

        free(threads);

        return 0;
    }

    thread_worker(loop);

    return 0;
}

void thread_worker(void *data) {
    uv_loop_t *loop = data;
    while (working) {
        uv_mutex_lock(&loop_mutex);
        uv_run(loop, UV_RUN_ONCE);
        uv_mutex_unlock(&loop_mutex);
    }
}

void api_connection_open(api_connection_t *api_connection) {
    api_connection->hints.ai_family = PF_INET;
    api_connection->hints.ai_socktype = SOCK_STREAM;
    api_connection->hints.ai_protocol = IPPROTO_TCP;
    api_connection->hints.ai_flags = 0;

    api_connection->resolver.data = api_connection;

    int err;
    if (ai.proxy_address_arg) {
        err = uv_getaddrinfo(loop, &api_connection->resolver, on_resolved, ai.proxy_address_arg, ai.proxy_port_arg,
                             &api_connection->hints);
    } else {
        err = uv_getaddrinfo(loop, &api_connection->resolver, on_resolved, ai.remote_address_arg, ai.remote_port_arg,
                             &api_connection->hints);
    }
    if (err) {
        fprintf(stderr, "getaddrinfo call error %s\n", uv_err_name(err));
        memset(api_connection, 0, offsetof(api_connection_t, keep_alive));
        LIST_INSERT_HEAD(&unconnected_head, api_connection, unconnected);
    }
}

void tcp_new_connection_cb(uv_stream_t *stream, int status) {
    if (status < 0) {
        fprintf(stderr, "Error on connection: %s.\n", uv_strerror(status));
        return;
    }

    http_request_t *http_request = calloc(1, sizeof(http_request_t));

    uv_tcp_init(loop, (uv_tcp_t *) &http_request->stream);

    http_request->stream.data = http_request;
    http_request->req.data = http_request;

    if (uv_accept(stream, &http_request->stream) == 0) {
        uv_read_start(&http_request->stream, alloc_buffer, tcp_read_cb);
    } else {
        uv_close((uv_handle_t *) &http_request->stream, NULL);
    }
}

void tcp_read_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf) {
    http_request_t *http_request = stream->data;

    int i;

    if (ai.debug_flag) {
        fprintf(stderr, "Read %zd\n", nread);
    }

    if (nread < 0) {
        if (nread != UV_EOF) {
            fprintf(stderr, "Error on reading: %s.\n", uv_strerror(nread));
        }

        http_request->keep_alive = 0;
        cleanup(http_request);
        return;
    }

    if (nread == 0) { //read nothing, skip it
        return;
    }

    // apache modules send '\0' after request, just ignore it
    if (nread == 1 && *buf->base == '\0') {
        return;
    }

    if (http_request->message == NULL) {
        uv_mutex_lock(&messages_mutex);
        http_request->message = LIST_FIRST(&free_messages_head);

        if (ai.debug_flag) {
            fprintf(stderr, "Get message from free_messages %p\n", http_request->message);
        }

        if (http_request->message == NULL) {
            http_request->message = TAILQ_LAST(&used_messages_head, used_messages_list);

            if (ai.debug_flag) {
                fprintf(stderr, "Get message from used_messages %p\n", http_request->message);
            }

            if (http_request->message == NULL) {
                fprintf(stderr, "Hasn't messages.\n");
                uv_mutex_unlock(&messages_mutex);
                return;
            }

            if (http_request->message->request) {
                fast_response(http_request->message->request);
            }

            TAILQ_REMOVE(&used_messages_head, http_request->message, used_messages);

            for (i = 0; i < MAX_HTTP_BUFFERS; i++) {
                if (http_request->message->buffers[i].base && http_request->message->buffers[i].len) {
                    if (http_request->message->full_body != http_request->message->buffers[i].base) {
                        free(http_request->message->buffers[i].base);
                    }
                    http_request->message->buffers[i].base = 0;
                    http_request->message->buffers[i].len = 0;
                }
            }

        } else {
            LIST_REMOVE(http_request->message, free_messages);
        }
        uv_mutex_unlock(&messages_mutex);

        memset(http_request->message, 0, sizeof(message_t));

        http_request->message->request = http_request;
        http_request->message->parser.data = http_request;
        http_parser_init(&http_request->message->parser, HTTP_REQUEST);
    }

    if (nread > 0 && http_request->message->last_buffer < MAX_HTTP_BUFFERS - 1) {
        http_request->message->buffers[http_request->message->last_buffer] = *buf;
        http_request->message->buffers[http_request->message->last_buffer].len = nread;
        http_request->message->last_buffer++;
    }

    ssize_t parsed = http_parser_execute(&http_request->message->parser, &settings,
                                     buf->base, nread);

    if (parsed < nread) {
        fprintf(stderr, "Error on parsing HTTP request, parsed: %zd, read: %zd, err: %s\n",
                parsed, nread, http_errno_name(http_request->message->parser.http_errno));
        fprintf(stderr, "Request: %.*s\n", (int)nread, buf->base);

        http_request->keep_alive = 0;
        fast_response(http_request);
        return;
    }

    if (!http_request->body_is_final) {
        if (ai.debug_flag) {
            fprintf(stderr, "Read part of request, wait next part\n");
        }
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "Request has been read, process it!\n");
    }

    http_request->message->full_body_len = 0;
    for (i = 0; i < http_request->message->last_body; i++) {
        http_request->message->full_body_len += http_request->message->bodies[i].length;
    }

    if (http_request->message->full_body_len + 1 + sizeof("&ClientID=") - 1 + CLIENT_ID_LENGTH > MAX_FULL_BODY) {
        return;
    }

    http_request->message->full_body_len = 0;
    for (i = 0; i < http_request->message->last_body; i++) {
        memmove(http_request->message->full_body + http_request->message->full_body_len,
                http_request->message->bodies[i].chunk, http_request->message->bodies[i].length);
        http_request->message->full_body_len += http_request->message->bodies[i].length;
    }
    http_request->message->full_body[http_request->message->full_body_len] = '\0';

    char *client_id = strcasestr(http_request->message->full_body, "ClientID=");

    if (ai.debug_flag) {
        if (client_id) {
            fprintf(stderr, "ClientID from request: %s\n", client_id);
        } else {
            fprintf(stderr, "Request hasn't ClientID: %s\n", http_request->message->full_body);
        }
    }

    if (client_id == NULL
        || (client_id - http_request->message->full_body + sizeof("ClientID=") - 1) >= http_request->message->full_body_len
        || (*(client_id + sizeof("ClientID=") - 1) == '&')) {

        if (client_id) {
            char *p = client_id + sizeof("ClientID=") - 1;
            if (p - http_request->message->full_body < http_request->message->full_body_len && *p == '&') {
                p++;
            }
            memmove(client_id, p, http_request->message->full_body_len - (p - http_request->message->full_body));
            http_request->message->full_body_len -= p - client_id;
        }

        memmove(http_request->message->full_body + http_request->message->full_body_len, "&ClientID=", sizeof("&ClientID=") - 1);
        http_request->message->full_body_len += sizeof("&ClientID=") - 1;

        http_request->message->new_client_id = http_request->message->full_body + http_request->message->full_body_len;

        for (i = 0; i < CLIENT_ID_LENGTH; i++) {
            int n = rand() % 32;
            if (n < 10) {
                http_request->message->new_client_id[i] = '0' + n;
            } else {
                http_request->message->new_client_id[i] = 'a' + n - 10;
            }
        }

        http_request->message->new_client_id[CLIENT_ID_LENGTH] = '\0';
        http_request->message->full_body_len += CLIENT_ID_LENGTH;
    }

    if (ai.remote_timeout_arg > 0) {
        uv_timer_init(loop, &http_request->timer);
        http_request->timer.data = http_request;
        uv_timer_start(&http_request->timer, time_is_up, ai.remote_timeout_arg, 0);
    }

    uv_mutex_lock(&messages_mutex);
    if (ai.debug_flag) {
        fprintf(stderr, "http_request->message->used_messages.tqe_next %p\n", http_request->message->used_messages.tqe_next);
        fprintf(stderr, "http_request->message->used_messages.tqe_prev %p\n", http_request->message->used_messages.tqe_prev);
        fprintf(stderr, "used_messages_head %p\n", &used_messages_head);
    }
    TAILQ_INSERT_HEAD(&used_messages_head, http_request->message, used_messages);
    if (ai.debug_flag) {
        fprintf(stderr, "http_request->message->used_messages.tqe_next %p\n", http_request->message->used_messages.tqe_next);
        fprintf(stderr, "http_request->message->used_messages.tqe_prev %p\n", http_request->message->used_messages.tqe_prev);
        fprintf(stderr, "used_messages_head %p\n", &used_messages_head);
    }
    uv_mutex_unlock(&messages_mutex);

    uv_async_send(&send_message);

    if (ai.async_flag) {
        fast_response(http_request);
    }
}

void tcp_write_cb(uv_write_t *req, int status) {
    http_request_t *http_request = req->data;

    if (http_request->closing) {
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "tcp_write_cb request %p, status %d\n", http_request, status);
    }

    if (http_request->stream.write_queue_size > 0) {
        if (ai.debug_flag) {
            fprintf(stderr, "write_queue_size == %zu, waiting when all response will be send\n", http_request->stream.write_queue_size);
        }
        return;
    }

    if (status == 0) {
        http_request->keep_alive = 0;
    }

    cleanup(http_request);

    if (status < 0) {
        fprintf(stderr, "Error on local connection: %s.\n", uv_strerror(status));
        return;
    }
}

int http_message_begin_cb(http_parser *parser) {
    http_request_t *http_request = parser->data;

    http_request->message->header_lines = 0;

    return 0;
}

int http_url_cb(http_parser *parser, const char *chunk, size_t len) {
    http_request_t *http_request = parser->data;

    http_request->message->url = chunk;
    http_request->message->url_len = len;

    return 0;
}

int http_header_field_cb(http_parser *parser, const char *chunk, size_t len) {
    http_request_t *http_request = parser->data;

    http_header_t *header = &http_request->message->headers[http_request->message->header_lines];

    header->field = chunk;
    header->field_length = len;

    return 0;
}

int http_header_value_cb(http_parser *parser, const char *chunk, size_t len) {
    http_request_t *http_request = parser->data;

    if (http_request->message->header_lines > MAX_HTTP_HEADERS) {
        fprintf(stderr, "Request contains to many HTTP headers\n");
        return 0;
    }

    http_header_t *header = &http_request->message->headers[http_request->message->header_lines];

    header->value = chunk;
    header->value_length = len;

    http_request->message->header_lines++;

    return 0;
}

int http_headers_complete_cb(http_parser *parser) {
    http_request_t *http_request = parser->data;

    http_request->keep_alive = http_should_keep_alive(&http_request->message->parser);

    if (ai.debug_flag) {
        fprintf(stderr, "HTTP connection keep-alive is %d\n", http_request->keep_alive);
        fprintf(stderr, "HTTP connection Content-Length: %llu\n", http_request->message->parser.content_length);
    }

    http_request->content_length = http_request->message->parser.content_length;

    if (http_request->content_length == ULLONG_MAX) {
        http_request->content_length = 0;
    }

    if (http_request->content_length > 0) {
        return 0;
    }

    http_request->body_is_final = 1;

    return 1;
}

int http_body_cb(http_parser *parser, const char *chunk, size_t len) {
    http_request_t *http_request = parser->data;

    http_request->message->bodies[http_request->message->last_body].chunk = chunk;
    http_request->message->bodies[http_request->message->last_body].length = len;

    http_request->message->last_body++;

    len = 0;
    int i;
    for (i = 0; i < http_request->message->last_body; i++) {
        len += http_request->message->bodies[i].length;
    }

    http_request->body_is_final = (len == http_request->content_length);

    if (ai.debug_flag) {
        fprintf(stderr, "Fetch next part of body, fetched %zu, is final: %d\n", len, http_request->body_is_final);
    }

    return 0;
}

void on_tcp_tls_read(uv_stream_t *stream, ssize_t nrd, const uv_buf_t *data)
{
    uv_tls_t *parent = CONTAINER_OF(stream, uv_tls_t, skt);
    if ( nrd <= 0 ) {
        if( nrd == UV_EOF) {
            uv_tls_close((uv_handle_t*)stream, api_connection_tls_closed);
        }
        free(data->base);
        ((uv_buf_t *) data)->base = NULL;
        return;
    }

    evt_tls_feed_data(parent->tls, data->base, nrd);
    free(data->base);
    ((uv_buf_t *) data)->base = NULL;
}

void on_resolved(uv_getaddrinfo_t *resolver, int status, struct addrinfo *res) {
    api_connection_t *api_connection = resolver->data;

    if (res && res->ai_addr == NULL) {
        goto error;
    }

    if (status < 0) {
        fprintf(stderr, "getaddrinfo callback error %s\n", uv_err_name(status));
        goto error;
    }

    api_connection->stream.data = api_connection;

    if (ai.ssl_flag) {
        uv_mutex_lock(&ssl_ctx_mutex);
        if (uv_tls_init(loop, &ssl_ctx, &api_connection->sclient) < 0 ) {
            uv_mutex_unlock(&ssl_ctx_mutex);
            fprintf(stderr, "uv_tls_init error\n");
            goto error;
        }
        uv_mutex_unlock(&ssl_ctx_mutex);

        api_connection->sclient.skt.data = api_connection;

        if (ai.proxy_address_arg) {
            uv_tcp_connect(&api_connection->cr, &api_connection->sclient.skt,
                           (const struct sockaddr *) res->ai_addr, on_remote_proxy_tls_connection);
        } else {
            uv_tcp_connect(&api_connection->cr, &api_connection->sclient.skt,
                           (const struct sockaddr *) res->ai_addr, on_remote_tls_connection);
        }
    } else {
        uv_tcp_init(loop, (uv_tcp_t *) &api_connection->stream);

        uv_tcp_connect(&api_connection->cr, (uv_tcp_t *) &api_connection->stream,
                       (const struct sockaddr *) res->ai_addr, on_remote_connection);
    }

    uv_freeaddrinfo(res);
    return;

error:
    uv_mutex_lock(&connections_mutex);
    LIST_INSERT_HEAD(&unconnected_head, api_connection, unconnected);
    uv_mutex_unlock(&connections_mutex);
    uv_freeaddrinfo(res);
    return;
}

void on_remote_proxy_tls_connection(uv_connect_t *req, int status) {
    api_connection_t *api_connection = req->handle->data;

    if (status < 0) {
        fprintf(stderr, "Error on remote connection: %s.\n", uv_strerror(status));
        api_connection_close(api_connection);
        return;
    }

    uv_read_start((uv_stream_t *) &(api_connection->sclient.skt), alloc_buffer,  on_remote_proxy_tls_connection_read);

    uv_buf_t *buf = &api_connection->buf;

    buf->len = sizeof("CONNECT ") - 1;
    buf->len += strlen(ai.remote_address_arg);
    buf->len += sizeof(":") - 1;
    buf->len += strlen(ai.remote_port_arg);
    buf->len += sizeof(" HTTP/1.1\r\n") - 1;
    buf->len += sizeof("Host: ") - 1;
    buf->len += strlen(ai.remote_address_arg);
    buf->len += sizeof(":") - 1;
    buf->len += strlen(ai.remote_port_arg);
    buf->len += sizeof(RN) - 1;
    buf->len += sizeof(RN);

    alloc_buffer(NULL, buf->len, buf);

    if (buf->base == NULL) {
        fprintf(stderr, "alloc_buffer failed\n");
        api_connection_close(api_connection);
        return;
    }

    buf->len = snprintf(buf->base, buf->len, "CONNECT %s:%s HTTP/1.1\r\nHost: %s:%s\r\n\r\n",
                        ai.remote_address_arg, ai.remote_port_arg, ai.remote_address_arg, ai.remote_port_arg);

    uv_write(&api_connection->req, (uv_stream_t *) &api_connection->sclient.skt, buf, 1, NULL);
}

void on_remote_proxy_tls_connection_read(uv_stream_t *stream, ssize_t nrd, const uv_buf_t *data)
{
    uv_tls_t *parent = CONTAINER_OF(stream, uv_tls_t, skt);

    api_connection_t *api_connection = parent->skt.data;

    free(api_connection->buf.base);
    api_connection->buf.base = 0;
    api_connection->buf.len = 0;

    if (nrd < sizeof("HTTP/1.1 200") - 1 || strncmp(data->base, "HTTP/1.1 200", sizeof("HTTP/1.1 200") - 1)) {
        fprintf(stderr, "Proxy server returns not 200 response\n");
        free(data->base);
        ((uv_buf_t *) data)->base = NULL;
        api_connection_close(api_connection);
        return;
    }

    free(data->base);
    ((uv_buf_t *) data)->base = NULL;

    uv_read_stop((uv_stream_t *) &(api_connection->sclient.skt));

    on_remote_tls_connection(&api_connection->cr, 0);
}

void on_remote_tls_connection(uv_connect_t *req, int status) {
    api_connection_t *api_connection = req->handle->data;

    if (status < 0) {
        fprintf(stderr, "Error on remote connection: %s.\n", uv_strerror(status));
        api_connection_close(api_connection);
        return;
    }

    uv_tls_connect(&api_connection->sclient, on_remote_tls_handshake);
    uv_read_stop((uv_stream_t *) &(api_connection->sclient.skt));
    uv_read_start((uv_stream_t *) &(api_connection->sclient.skt), alloc_buffer, on_tcp_tls_read);
}

void on_remote_tls_handshake(uv_tls_t *ut, int status) {
    api_connection_t *api_connection = ut->skt.data;

    if (status < 0) {
        fprintf(stderr, "Error on remote TLS connection.");
        api_connection_close(api_connection);
        return;
    }

    on_remote_connection(&api_connection->cr, 0);
}

void on_remote_connection(uv_connect_t *req, int status) {
    api_connection_t *api_connection = req->handle->data;

    if (status < 0) {
        fprintf(stderr, "Error on remote connection: %s.\n", uv_strerror(status));
        api_connection_close(api_connection);
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "Connection to %s:%s is established\n", ai.remote_address_arg, ai.remote_port_arg);
    }

    api_connection->is_connected = 1;
    uv_mutex_lock(&connections_mutex);
    LIST_INSERT_HEAD(&connected_head, api_connection, connected);
    uv_mutex_unlock(&connections_mutex);

    // flush the queue if it existed
    uv_async_send(&send_message);
}

void send_message_cb(uv_async_t* handle) {
    uv_mutex_lock(&messages_mutex);

    if (ai.debug_flag) {
        fprintf(stderr, "Send message cb\n");
    }

    message_t *message = TAILQ_LAST(&used_messages_head, used_messages_list);
    if (message == NULL) {
        uv_mutex_unlock(&messages_mutex);
        if (ai.debug_flag) {
            fprintf(stderr, "Send message cb haven't message in queue\n");
        }
        return;
    }

    uv_mutex_lock(&connections_mutex);
    api_connection_t *api_connection = LIST_FIRST(&connected_head);

    if (api_connection == NULL) {
        uv_mutex_unlock(&connections_mutex);
        uv_mutex_unlock(&messages_mutex);
        if (ai.debug_flag) {
            fprintf(stderr, "Send message cb haven't free connection\n");
        }
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "Send message cb process message %p over connection %p\n", message, api_connection);
        fprintf(stderr, "message->used_messages.tqe_next %p\n", message->used_messages.tqe_next);
        fprintf(stderr, "message->used_messages.tqe_prev %p\n", message->used_messages.tqe_prev);
        fprintf(stderr, "used_messages_head %p\n", &used_messages_head);
    }

    TAILQ_REMOVE(&used_messages_head, message, used_messages);

    if (ai.debug_flag) {
        fprintf(stderr, "message->used_messages.tqe_next %p\n", message->used_messages.tqe_next);
        fprintf(stderr, "message->used_messages.tqe_prev %p\n", message->used_messages.tqe_prev);
        fprintf(stderr, "used_messages_head %p\n", &used_messages_head);
    }

    if (ai.debug_flag) {
        message_t *m;
        fprintf(stderr, "Used messages after remove message %p has:\n", message);

        TAILQ_FOREACH(m, &used_messages_head, used_messages) {
            fprintf(stderr, "Message %p\n", m);
        }
    }

    uv_mutex_unlock(&messages_mutex);

    LIST_REMOVE(api_connection, connected);
    uv_mutex_unlock(&connections_mutex);

    api_connection->message = message;
    send_to_remote(api_connection);
}

void send_to_remote(api_connection_t *api_connection) {
    if (api_connection->message->last_buffer > 0) {

        if (((ai.proxy_address_arg == NULL || !ai.ssl_flag) && api_connection->message->new_client_id == NULL)
            || api_connection->message->last_buffer + 2 > MAX_HTTP_BUFFERS) {
            if (ai.ssl_flag) {
                api_connection->send_last_buffer = 0;
                uv_tls_write(&api_connection->sclient, api_connection->message->buffers, tcp_write_tls_remote_cb);
            } else {
                uv_write(&api_connection->req, &api_connection->stream,
                         api_connection->message->buffers, api_connection->message->last_buffer, tcp_write_remote_cb);
            }
        } else {
            uv_buf_t *buf = &api_connection->message->buffers[api_connection->message->last_buffer];
            uv_buf_t *body = &api_connection->message->buffers[api_connection->message->last_buffer + 1];

            api_connection->message->last_buffer += 2;

            const char *method = http_method_str(api_connection->message->parser.method);

            buf->len = strlen(method);
            buf->len += sizeof(" ") - 1;
            if (ai.proxy_address_arg && !ai.ssl_flag) {
                buf->len += sizeof("http://") - 1;
                buf->len += strlen(ai.remote_address_arg);
                buf->len += sizeof(":") - 1;
                buf->len += strlen(ai.remote_port_arg);
            }
            buf->len += api_connection->message->url_len;
            buf->len += sizeof(" HTTP/1.1") - 1;
            buf->len += sizeof(RN) - 1;

            int i;
            for (i = 0; i < api_connection->message->header_lines; i++) {
                if (api_connection->message->headers[i].field_length != sizeof("Content-Length") - 1
                    || strncmp(api_connection->message->headers[i].field, "Content-Length", sizeof("Content-Length") - 1) != 0) {
                    if (api_connection->message->headers[i].field_length > 0) {
                        buf->len += api_connection->message->headers[i].field_length;
                        buf->len += sizeof(": ") - 1;
                    }
                    if (api_connection->message->headers[i].field_length == 0 && i > 0 && api_connection->message->headers[i - 1].field_length > 0){
                        buf->len -= sizeof(RN) - 1;
                    }
                    buf->len += api_connection->message->headers[i].value_length;
                    buf->len += sizeof(RN) - 1;
                }
            }

            buf->len += sizeof("Content-Length: 2147483647") - 1;
            buf->len += sizeof(RN) - 1;

            buf->len += sizeof(RN) - 1;

            buf->base = malloc(buf->len);
            if (buf->base == NULL) {
                return;
            }

            buf->len = 0;

            memmove(buf->base + buf->len, method, strlen(method));
            buf->len += strlen(method);

            memmove(buf->base + buf->len, " ", sizeof(" ") - 1);
            buf->len += sizeof(" ") - 1;

            if (ai.proxy_address_arg && !ai.ssl_flag) {
                buf->len += sprintf(buf->base + buf->len, "http://%s:%s", ai.remote_address_arg, ai.remote_port_arg);
            }

            memmove(buf->base + buf->len, api_connection->message->url, api_connection->message->url_len);
            buf->len += api_connection->message->url_len;

            memmove(buf->base + buf->len, " HTTP/1.1", sizeof(" HTTP/1.1") - 1);
            buf->len += sizeof(" HTTP/1.1") - 1;

            memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
            buf->len += sizeof(RN) - 1;

            for (i = 0; i < api_connection->message->header_lines; i++) {
                if (api_connection->message->headers[i].field_length != sizeof("Content-Length") - 1
                    || strncmp(api_connection->message->headers[i].field, "Content-Length", sizeof("Content-Length") - 1) != 0) {
                    if (api_connection->message->headers[i].field_length > 0) {
                        memmove(buf->base + buf->len, api_connection->message->headers[i].field,
                                api_connection->message->headers[i].field_length);
                        buf->len += api_connection->message->headers[i].field_length;

                        memmove(buf->base + buf->len, ": ", sizeof(": ") - 1);
                        buf->len += sizeof(": ") - 1;
                    }

                    if (api_connection->message->headers[i].field_length == 0 && i > 0 && api_connection->message->headers[i - 1].field_length > 0){
                        buf->len -= sizeof(RN) - 1;
                    }

                    memmove(buf->base + buf->len, api_connection->message->headers[i].value,
                            api_connection->message->headers[i].value_length);
                    buf->len += api_connection->message->headers[i].value_length;

                    memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
                    buf->len += sizeof(RN) - 1;
                }
            }

            buf->len += sprintf(buf->base + buf->len, "Content-Length: %zu", api_connection->message->full_body_len);

            memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
            buf->len += sizeof(RN) - 1;

            memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
            buf->len += sizeof(RN) - 1;

            body->base = api_connection->message->full_body;
            body->len = api_connection->message->full_body_len;

            if (ai.ssl_flag) {
                api_connection->send_last_buffer = api_connection->message->last_buffer - 2;
                uv_tls_write(&api_connection->sclient, buf, tcp_write_tls_remote_cb);
            } else {
                uv_write(&api_connection->req, &api_connection->stream, buf, 2,
                         tcp_write_remote_cb);
            }
        }

        http_parser_init(&api_connection->message->parser, HTTP_RESPONSE);
        api_connection->message->parser.data = api_connection;

        if (ai.ssl_flag) {
            uv_tls_read((uv_stream_t *) &api_connection->sclient, alloc_buffer, tcp_read_tls_remote_cb);
        } else {
            uv_read_start(api_connection->req.handle, alloc_buffer, tcp_read_remote_cb);
        }
    }
}

void tcp_write_tls_remote_cb(uv_tls_t *tls, int status) {
    api_connection_t *api_connection = tls->skt.data;

    if (status < 0) {
        fprintf(stderr, "Error on remote SSL connection.\n");
        api_connection_close(api_connection);
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "tcp_write_tls_remote_cb: status %d\n", status);
    }

    if (status >= api_connection->message->buffers[api_connection->send_last_buffer].len) {
        api_connection->send_last_buffer++;
    } else {
        api_connection->message->buffers[api_connection->send_last_buffer].len -= status;
        api_connection->message->buffers[api_connection->send_last_buffer].base += status;
    }

    if (api_connection->send_last_buffer < api_connection->message->last_buffer) {
        uv_tls_write(&api_connection->sclient, &api_connection->message->buffers[api_connection->send_last_buffer], tcp_write_tls_remote_cb);
    } else {
        tcp_write_remote(api_connection);
    }
}

void tcp_write_remote_cb(uv_write_t *req, int status) {
    api_connection_t *api_connection = req->handle->data;

    if (status < 0) {
        fprintf(stderr, "Error on remote connection: %s.\n", uv_strerror(status));
        api_connection_close(api_connection);
        return;
    }

    tcp_write_remote(api_connection);
}

void tcp_write_remote(api_connection_t *api_connection) {
    if (ai.debug_flag) {
        fprintf(stderr, "Sent message %p over connection %p\n", api_connection->message, api_connection);
    }


    int i;
    for (i = 0; i < MAX_HTTP_BUFFERS; i++) {
        if (api_connection->message->buffers[i].base && api_connection->message->buffers[i].len) {
            if (api_connection->message->full_body != api_connection->message->buffers[i].base) {
                free(api_connection->message->buffers[i].base);
            }
            api_connection->message->buffers[i].base = 0;
            api_connection->message->buffers[i].len = 0;
        }
    }

    api_connection->message->last_buffer = 0;

    for (i = 0; i < MAX_HTTP_HEADERS; i++) {
        api_connection->message->headers[i].field = 0;
        api_connection->message->headers[i].field_length = 0;
        api_connection->message->headers[i].value = 0;
        api_connection->message->headers[i].value_length = 0;
    }

    api_connection->message->header_lines = 0;

    for (i = 0; i < MAX_HTTP_BODIES; i++) {
        api_connection->message->bodies[i].chunk = 0;
        api_connection->message->bodies[i].length = 0;
    }

    api_connection->message->last_body = 0;
}

void tcp_read_tls_remote_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf) {
    uv_tls_t *tls = (uv_tls_t *) stream;

    api_connection_t *api_connection = tls->skt.data;

    //evt_tls doesn't use alloc_cb, well do it by hand;
    uv_buf_t nb;
    alloc_buffer(NULL, buf->len, &nb);
    memmove(nb.base, buf->base, buf->len);

    tcp_read_remote(api_connection, nread, &nb);
}

void tcp_read_remote_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf) {
    api_connection_t *api_connection = stream->data;

    if (nread < 0) {
        if (nread != UV_EOF) {
            fprintf(stderr, "Error on reading: %s.\n", uv_strerror(nread));
        } else {
            fprintf(stderr, "read EOF\n");
        }
        api_connection_close(api_connection);
        return;
    }

    tcp_read_remote(api_connection, nread, buf);
}

static char *
_strnstr(const char *s1, const char *s2, size_t len)
{
	size_t l2;

	l2 = strlen(s2);
	if (!l2)
		return (char *)s1;
	while (len >= l2) {
		len--;
		if (!memcmp(s1, s2, l2))
			return (char *)s1;
		s1++;
	}
	return NULL;
}


void tcp_read_remote(api_connection_t *api_connection, ssize_t nread, const uv_buf_t *buf) {
    if (ai.debug_flag) {
        fprintf(stderr, "Read from API server: %zd\n", nread);
        fprintf(stderr, "Read from API server request: %s\n", buf->base);
    }

    if (api_connection->message == NULL) {
        return;
    }

    if (nread > 0 && api_connection->message->last_buffer < MAX_HTTP_BUFFERS - 1) {
        api_connection->message->buffers[api_connection->message->last_buffer] = *buf;
        api_connection->message->buffers[api_connection->message->last_buffer].len = nread;
        api_connection->message->last_buffer++;
    }

    ssize_t parsed = http_parser_execute(&api_connection->message->parser, &api_settings,
                                     buf->base, nread);

    if (parsed < nread) {
        fprintf(stderr, "Error on parsing API server HTTP response, parsed: %zd, read: %zd, err: %s\n",
                parsed, nread, http_errno_name(api_connection->message->parser.http_errno));
        fprintf(stderr, "Request: %.*s\n", (int)nread, buf->base);

        api_connection_close(api_connection);
        return;
    }

    if (!api_connection->body_is_final) {
        if (ai.debug_flag) {
            fprintf(stderr, "Read part of API response, wait next part\n");
        }
        return;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "Response has been read, finish it!\n");
    }

    if (api_connection->message->request) {
        if (ai.debug_flag) {
            fprintf(stderr, "API server connection has request, send response back, new_client_id: %p\n",
                api_connection->message->new_client_id);
        }
        if (api_connection->message->new_client_id == NULL || api_connection->message->last_buffer + 2 > MAX_HTTP_BUFFERS) {
            uv_write(&api_connection->message->request->req, &api_connection->message->request->stream,
                     api_connection->message->buffers, api_connection->message->last_buffer, tcp_write_cb);
        } else {
            uv_buf_t *buf = &api_connection->message->buffers[api_connection->message->last_buffer];
            uv_buf_t *body = &api_connection->message->buffers[api_connection->message->last_buffer + 1];

            api_connection->message->last_buffer += 2;

            unsigned int status = api_connection->message->parser.status_code;

            buf->len = 0;

            switch (status) {
                case 200:
                buf->len += sizeof("HTTP/1.1 200 OK") - 1;
                break;

                case 301:
                buf->len += sizeof("HTTP/1.1 301 Moved Permanently") - 1;
                break;

                case 302:
                buf->len += sizeof("HTTP/1.1 302 Found") - 1;
                break;

                case 403:
                buf->len += sizeof("HTTP/1.1 403 Forbidden") - 1;
                break;

                default:
                buf->len += sizeof("HTTP/1.1 XXX Another") -1;
            }

            buf->len += sizeof(RN) - 1;

            int i;
            for (i = 0; i < api_connection->message->header_lines; i++) {
                buf->len += api_connection->message->headers[i].field_length;
                buf->len += sizeof(": ") - 1;

                if (api_connection->message->headers[i].field_length == 0 && i > 0 && api_connection->message->headers[i - 1].field_length > 0){
                    buf->len -= sizeof(RN) - 1;
                }

                buf->len += api_connection->message->headers[i].value_length;
                buf->len += sizeof(RN) - 1;

                if (api_connection->message->headers[i].field_length == sizeof("X-DataDome-headers") - 1
                    && strncmp(api_connection->message->headers[i].field, "X-DataDome-headers", sizeof("X-DataDome-headers") - 1) == 0) {
                    buf->len += sizeof(" Set-Cookie ") - 1;
                }
            }

            buf->len += sizeof("Set-Cookie: ") - 1;
            buf->len += strlen(ai.client_id_arg);
            buf->len += sizeof("=") - 1;
            buf->len += CLIENT_ID_LENGTH;
            buf->len += sizeof(RN) - 1;

            buf->len += sizeof(RN) - 1;

            buf->base = malloc(buf->len);
            if (buf->base == NULL) {
                return;
            }

            buf->len = 0;

            switch (status) {
                case 200:
                memmove(buf->base + buf->len, "HTTP/1.1 200 OK", sizeof("HTTP/1.1 200 OK") - 1);
                buf->len += sizeof("HTTP/1.1 200 OK") - 1;
                break;

                case 301:
                memmove(buf->base + buf->len, "HTTP/1.1 301 Moved Permanently", sizeof("HTTP/1.1 301 Moved Permanently") - 1);
                buf->len += sizeof("HTTP/1.1 301 Moved Permanently") - 1;
                break;

                case 302:
                memmove(buf->base + buf->len, "HTTP/1.1 302 Found", sizeof("HTTP/1.1 302 Found") - 1);
                buf->len += sizeof("HTTP/1.1 302 Found") - 1;
                break;

                case 403:
                memmove(buf->base + buf->len, "HTTP/1.1 403 Forbidden", sizeof("HTTP/1.1 403 Forbidden") - 1);
                buf->len += sizeof("HTTP/1.1 403 Forbidden") - 1;
                break;

                default:
                snprintf(buf->base + buf->len, sizeof("HTTP/1.1 XXX Another"), "HTTP/1.1 %d Another", status);
                buf->len += sizeof("HTTP/1.1 XXX Another") - 1;
            }

            memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
            buf->len += sizeof(RN) - 1;

            int set_cookie = 0;

            for (i = 0; i < api_connection->message->header_lines; i++) {
                memmove(buf->base + buf->len, api_connection->message->headers[i].field,
                    api_connection->message->headers[i].field_length);
                buf->len += api_connection->message->headers[i].field_length;

                memmove(buf->base + buf->len, ": ", sizeof(": ") - 1);
                buf->len += sizeof(": ") - 1;

                if (api_connection->message->headers[i].field_length == sizeof("Set-Cookie") - 1
                    && strncmp(api_connection->message->headers[i].field, "Set-Cookie", sizeof("Set-Cookie") - 1) == 0) {
                    set_cookie = 1;

                    memmove(buf->base + buf->len, ai.client_id_arg, strlen(ai.client_id_arg));
                    buf->len += strlen(ai.client_id_arg);

                    memmove(buf->base + buf->len, "=", sizeof("=") - 1);
                    buf->len += sizeof("=") - 1;

                    memmove(buf->base + buf->len, api_connection->message->new_client_id, CLIENT_ID_LENGTH);
                    buf->len += CLIENT_ID_LENGTH;

                    memmove(buf->base + buf->len, ", ", sizeof(", ") - 1);
                    buf->len += sizeof(", ") - 1;
                }

                if (api_connection->message->headers[i].field_length == 0 && i > 0 && api_connection->message->headers[i - 1].field_length > 0){
                    buf->len -= sizeof(RN) - 1;
                }

                memmove(buf->base + buf->len, api_connection->message->headers[i].value,
                    api_connection->message->headers[i].value_length);
                buf->len += api_connection->message->headers[i].value_length;


                if (api_connection->message->headers[i].field_length == sizeof("X-DataDome-headers") - 1
                    && strncmp(api_connection->message->headers[i].field, "X-DataDome-headers", sizeof("X-DataDome-headers") - 1) == 0
                    && _strnstr(api_connection->message->headers[i].value, "Set-Cookie", api_connection->message->headers[i].value_length) == NULL) {
                memmove(buf->base + buf->len, api_connection->message->headers[i].value,
                    api_connection->message->headers[i].value_length);
                    memmove(buf->base + buf->len, " Set-Cookie ", sizeof(" Set-Cookie ") - 1);
                    buf->len += sizeof(" Set-Cookie ") - 1;
                }

                memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
                buf->len += sizeof(RN) - 1;
            }

            if (set_cookie == 0) {
                memmove(buf->base + buf->len, "Set-Cookie: ", sizeof("Set-Cookie: ") - 1);
                buf->len += sizeof("Set-Cookie: ") - 1;

                memmove(buf->base + buf->len, ai.client_id_arg, strlen(ai.client_id_arg));
                buf->len += strlen(ai.client_id_arg);

                memmove(buf->base + buf->len, "=", sizeof("=") - 1);
                buf->len += sizeof("=") - 1;

                memmove(buf->base + buf->len, api_connection->message->new_client_id, CLIENT_ID_LENGTH);
                buf->len += CLIENT_ID_LENGTH;

                memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
                buf->len += sizeof(RN) - 1;
            }

            memmove(buf->base + buf->len, RN, sizeof(RN) - 1);
            buf->len += sizeof(RN) - 1;

            api_connection->message->full_body_len = 0;

            for (i = 0; i < api_connection->message->last_body; i++) {
                memmove(api_connection->message->full_body + api_connection->message->full_body_len,
                    api_connection->message->bodies[i].chunk, api_connection->message->bodies[i].length);
                api_connection->message->full_body_len += api_connection->message->bodies[i].length;
            }

            body->base = api_connection->message->full_body;
            body->len = api_connection->message->full_body_len;

            uv_write(&api_connection->message->request->req, &api_connection->message->request->stream,
                     buf, body->len > 0 ? 2 : 1, tcp_write_cb);
        }
        api_connection->message->request->sent_response = 1;
    }


    api_cleanup(api_connection);

    if (api_connection->keep_alive) {
        api_connection->is_connected = 1;
        uv_mutex_lock(&connections_mutex);
        LIST_INSERT_HEAD(&connected_head, api_connection, connected);
        uv_mutex_unlock(&connections_mutex);

        // flush the queue if it existed
        uv_async_send(&send_message);
    } else {
        api_connection_close(api_connection);
    }
}


int api_http_message_begin_cb(http_parser *parser) {
    api_connection_t *api_connection = parser->data;

    api_connection->message->header_lines = 0;

    return 0;
}

int api_http_header_field_cb(http_parser *parser, const char *chunk, size_t len) {
    api_connection_t *api_connection = parser->data;

    http_header_t *header = &api_connection->message->headers[api_connection->message->header_lines];

    header->field = chunk;
    header->field_length = len;

    return 0;
}

int api_http_header_value_cb(http_parser *parser, const char *chunk, size_t len) {
    api_connection_t *api_connection = parser->data;

    if (api_connection->message->header_lines > MAX_HTTP_HEADERS) {
        fprintf(stderr, "Request contains to many HTTP headers\n");
        return 0;
    }

    http_header_t *header = &api_connection->message->headers[api_connection->message->header_lines];

    header->value = chunk;
    header->value_length = len;

    api_connection->message->header_lines++;

    return 0;
}

int api_http_headers_complete_cb(http_parser *parser) {
    api_connection_t *api_connection = parser->data;

    api_connection->keep_alive = http_should_keep_alive(&api_connection->message->parser);

    if (ai.debug_flag) {
        fprintf(stderr, "API server connection keep-alive is %d\n", api_connection->keep_alive);
        fprintf(stderr, "API server connection Content-Length: %llu\n", api_connection->message->parser.content_length);
    }

    api_connection->content_length = api_connection->message->parser.content_length;

    if (api_connection->content_length == ULLONG_MAX) {
        api_connection->content_length = 0;
    }

    if (api_connection->content_length > 0) {
        api_connection->body_is_final = 0;
        return 0;
    }

    api_connection->body_is_final = 1;

    return 1;
}

int api_http_body_cb(http_parser *parser, const char *chunk, size_t len) {
    api_connection_t *api_connection = parser->data;

    if (api_connection->body_is_final) {
        if (ai.debug_flag) {
            fprintf(stderr, "Full body fetched, skip this part\n");
        }
        return 0;
    }

    api_connection->message->bodies[api_connection->message->last_body].chunk = chunk;
    api_connection->message->bodies[api_connection->message->last_body].length = len;

    api_connection->message->last_body++;

    len = 0;
    int i;
    for (i = 0; i < api_connection->message->last_body; i++) {
        len += api_connection->message->bodies[i].length;
    }

    api_connection->body_is_final = (len == api_connection->content_length);

    if (ai.debug_flag) {
        fprintf(stderr, "Fetch next part of API response body, fetched %zu, is final: %d\n", len, api_connection->body_is_final);
    }

    return 0;
}

void time_is_up(uv_timer_t* handle) {
    http_request_t *http_request = handle->data;

    if (ai.debug_flag) {
        fprintf(stderr, "Time is up :( \n");
    }

    fast_response(http_request);
}

int fast_response(http_request_t *http_request) {
    if (http_request == NULL) {
        return 0;
    }

    if (ai.debug_flag) {
        fprintf(stderr, "Fast response for request %p, send_response: %d\n", http_request, http_request->sent_response);
    }

    if (http_request->sent_response) {
        return 0;
    }

    http_request->response.len = 0;
    if (ai.forbidden_flag) {
        http_request->response.len += sizeof(RESPONSE_FORBIDDEN) - 1;
    } else {
        http_request->response.len += sizeof(RESPONSE_OK) - 1;
    }

    http_request->response.len += sizeof(RN) - 1;

    if (http_request->keep_alive) {
        http_request->response.len += sizeof(RESPONSE_CONNECTION_KEEP_ALIVE) - 1;
    } else {
        http_request->response.len += sizeof(RESPONSE_CONNECTION_CLOSE) - 1;
    }

    http_request->response.len += sizeof(RN) - 1;

    http_request->response.len += sizeof(RESPONSE_CONTENT_LENGTH) - 1;
    http_request->response.len += sizeof(RN) - 1;

    http_request->response.len += sizeof(RESPONSE_SERVER) - 1;
    http_request->response.len += sizeof(RN) - 1;

    if (ai.forbidden_flag) {
        http_request->response.len += sizeof(RESPONSE_X_DataDomeResponse_FORBIDDEN) - 1;
        http_request->response.len += sizeof(RN) - 1;
    } else {
        http_request->response.len += sizeof(RESPONSE_X_DataDomeResponse_OK) - 1;
        http_request->response.len += sizeof(RN) - 1;
    }

    http_request->response.len += sizeof(RESPONSE_PRAGMA) - 1;
    http_request->response.len += sizeof(RN) - 1;

    http_request->response.len += sizeof(RESPONSE_X_DataDome) - 1;
    http_request->response.len += sizeof(RN) - 1;

    if (http_request->message && http_request->message->new_client_id) {
        http_request->response.len += sizeof("Set-Cookie: ") - 1;
        http_request->response.len += strlen(ai.client_id_arg);
        http_request->response.len += sizeof("=") - 1;
        http_request->response.len += CLIENT_ID_LENGTH;
        http_request->response.len += sizeof(RN) - 1;
    }

    http_request->response.len += sizeof(RESPONSE_X_DataDome_headers) - 1;
    if (http_request->message && http_request->message->new_client_id) {
        http_request->response.len += sizeof(" Set-Cookie") - 1;
    }
    http_request->response.len += sizeof(RN) - 1;

    http_request->response.len += sizeof(RN) - 1;

    http_request->response.base = malloc(http_request->response.len + 1);
    if (http_request->response.base == NULL) {
        http_request->keep_alive = 0;
        http_request->response.len = 0;
        cleanup(http_request);
        return -1;
    }

    size_t used = 0;
    if (ai.forbidden_flag) {
        memmove(http_request->response.base + used, RESPONSE_FORBIDDEN, sizeof(RESPONSE_FORBIDDEN) - 1);
        used += sizeof(RESPONSE_FORBIDDEN) - 1;
    } else {
        memmove(http_request->response.base + used, RESPONSE_OK, sizeof(RESPONSE_OK) - 1);
        used += sizeof(RESPONSE_OK) - 1;
    }

    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    if (http_request->keep_alive) {
        memmove(http_request->response.base + used, RESPONSE_CONNECTION_KEEP_ALIVE, sizeof(RESPONSE_CONNECTION_KEEP_ALIVE) - 1);
        used += sizeof(RESPONSE_CONNECTION_KEEP_ALIVE) - 1;
    } else {
        memmove(http_request->response.base + used, RESPONSE_CONNECTION_CLOSE, sizeof(RESPONSE_CONNECTION_CLOSE) - 1);
        used += sizeof(RESPONSE_CONNECTION_CLOSE) - 1;
    }

    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    memmove(http_request->response.base + used, RESPONSE_CONTENT_LENGTH, sizeof(RESPONSE_CONTENT_LENGTH) - 1);
    used += sizeof(RESPONSE_CONTENT_LENGTH) - 1;
    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    memmove(http_request->response.base + used, RESPONSE_SERVER, sizeof(RESPONSE_SERVER) - 1);
    used += sizeof(RESPONSE_SERVER) - 1;
    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    if (ai.forbidden_flag) {
        memmove(http_request->response.base + used, RESPONSE_X_DataDomeResponse_FORBIDDEN, sizeof(RESPONSE_X_DataDomeResponse_FORBIDDEN) - 1);
        used += sizeof(RESPONSE_X_DataDomeResponse_FORBIDDEN) - 1;
        memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
        used += sizeof(RN) - 1;
    } else {
        memmove(http_request->response.base + used, RESPONSE_X_DataDomeResponse_OK, sizeof(RESPONSE_X_DataDomeResponse_OK) - 1);
        used += sizeof(RESPONSE_X_DataDomeResponse_OK) - 1;
        memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
        used += sizeof(RN) - 1;
    }

    memmove(http_request->response.base + used, RESPONSE_PRAGMA, sizeof(RESPONSE_PRAGMA) - 1);
    used += sizeof(RESPONSE_PRAGMA) - 1;
    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    memmove(http_request->response.base + used, RESPONSE_X_DataDome, sizeof(RESPONSE_X_DataDome) - 1);
    used += sizeof(RESPONSE_X_DataDome) - 1;
    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    if (http_request->message && http_request->message->new_client_id) {
        memmove(http_request->response.base + used, "Set-Cookie: ", sizeof("Set-Cookie: ") - 1);
        used += sizeof("Set-Cookie: ") - 1;

        memmove(http_request->response.base + used, ai.client_id_arg, strlen(ai.client_id_arg));
        used += strlen(ai.client_id_arg);

        memmove(http_request->response.base + used, "=", sizeof("=") - 1);
        used += sizeof("=") - 1;

        memmove(http_request->response.base + used, http_request->message->new_client_id, CLIENT_ID_LENGTH);
        used += CLIENT_ID_LENGTH;

        memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
        used += sizeof(RN) - 1;
    }

    memmove(http_request->response.base + used, RESPONSE_X_DataDome_headers, sizeof(RESPONSE_X_DataDome_headers) - 1);
    used += sizeof(RESPONSE_X_DataDome_headers) - 1;
    if (http_request->message && http_request->message->new_client_id) {
        memmove(http_request->response.base + used, " Set-Cookie", sizeof(" Set-Cookie") - 1);
        used += sizeof(" Set-Cookie") - 1;
    }
    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    memmove(http_request->response.base + used, RN, sizeof(RN) - 1);
    used += sizeof(RN) - 1;

    if (http_request->message) {
        http_request->message->request = 0;
        http_request->message = 0;
    }

    uv_write(&http_request->req, &http_request->stream, &http_request->response, 1, tcp_write_cb);

    return 0;
}

void cleanup(http_request_t *http_request) {
    uv_timer_stop(&http_request->timer);

    if (ai.debug_flag) {
        fprintf(stderr, "Cleanup request %p, message %p \n", http_request, http_request->message);
    }

    if (http_request->message) {
        http_request->message->request = 0;
        http_request->message = 0;
    }

    if (http_request->response.base) {
        free(http_request->response.base);
        http_request->response.base = 0;
    }

    http_request->response.len = 0;
    http_request->response.base = 0;

    if (!http_request->keep_alive) {
        http_request->closing = 1;
        uv_close((uv_handle_t *) &http_request->stream, connection_closed);
    }
}

void reconnect_connection_timer_cb(uv_timer_t* handle) {
    api_connection_t *api_connection, *api_connection_temp;

    uv_mutex_lock(&connections_mutex);
    LIST_FOREACH(api_connection, &unconnected_head, unconnected) {
        if (ai.debug_flag) {
            fprintf(stderr, "Reconnect connection: %p\n", api_connection);
        }

        LIST_REMOVE(api_connection, unconnected);
        api_cleanup(api_connection);
        api_connection_open(api_connection);
    }
    uv_mutex_unlock(&connections_mutex);
}

void api_cleanup(api_connection_t *api_connection) {
    if (ai.debug_flag) {
        fprintf(stderr, "Cleanup API connection %p, message %p\n", api_connection, api_connection->message);
    }

    if (api_connection->message && api_connection->message->request) {
        fast_response(api_connection->message->request);
    }

    if (api_connection->message) {
        int i;
        for (i = 0; i < MAX_HTTP_BUFFERS; i++) {
            if (api_connection->message->buffers[i].base && api_connection->message->buffers[i].len) {
                if (api_connection->message->full_body != api_connection->message->buffers[i].base) {
                    free(api_connection->message->buffers[i].base);
                }
                api_connection->message->buffers[i].base = 0;
                api_connection->message->buffers[i].len = 0;
            }
        }

        api_connection->message->header_lines = 0;
        api_connection->message->last_buffer = 0;

        memset(api_connection->message, 0, sizeof(message_t));

        if (ai.debug_flag) {
            fprintf(stderr, "Put message to free_messages %p\n", api_connection->message);
        }

        uv_mutex_lock(&messages_mutex);
        LIST_INSERT_HEAD(&free_messages_head, api_connection->message, free_messages);
        uv_mutex_unlock(&messages_mutex);

        api_connection->message = 0;
    }

    api_connection->send_last_buffer = 0;
}

void api_connection_close(api_connection_t *api_connection) {
    if (api_connection->is_connected) {
        uv_mutex_lock(&connections_mutex);
        LIST_REMOVE(api_connection, connected);
        uv_mutex_unlock(&connections_mutex);
    }

    api_cleanup(api_connection);

    if (ai.debug_flag) {
        fprintf(stderr, "Connection to %s:%s is closing\n", ai.remote_address_arg, ai.remote_port_arg);
    }

    if (ai.ssl_flag) {
        uv_mutex_lock(&ssl_ctx_mutex);
        int r = uv_tls_close((uv_handle_t *) &api_connection->sclient, api_connection_tls_closed);
        if (r != 1) {
            if (ai.debug_flag) {
                fprintf(stderr, "Force close connection\n");
            }
            uv_tls_t *t = (uv_tls_t*)&api_connection->sclient;
            evt_tls_t *tls = t->tls;
            uv_tls_t *ut = (uv_tls_t*)tls->data;

            evt_tls_free(tls);
            if ( !uv_is_closing((uv_handle_t*)&(ut->skt)))
                uv_close( (uv_handle_t*)&(ut->skt), api_connection_tls_closed);
        }
        uv_mutex_unlock(&ssl_ctx_mutex);
    } else {
        uv_close((uv_handle_t *) &api_connection->stream, api_connection_closed);
    }
}

void connection_closed(uv_handle_t* handle) {
    http_request_t *http_request = handle->data;

    handle->data = NULL;
    free(http_request);
}

void api_connection_tls_closed(uv_handle_t* handle) {
    api_connection_t *api_connection = CONTAINER_OF(handle, api_connection_t, sclient);

    if (ai.debug_flag) {
        fprintf(stderr, "Close TLS connection: %p\n", api_connection);
    }

    memset(api_connection, 0, offsetof(api_connection_t, connected));

    uv_mutex_lock(&connections_mutex);
    LIST_INSERT_HEAD(&unconnected_head, api_connection, unconnected);
    uv_mutex_unlock(&connections_mutex);
}

void api_connection_closed(uv_handle_t* handle) {
    api_connection_t *api_connection = handle->data;

    if (ai.debug_flag) {
        fprintf(stderr, "Close connection: %p\n", api_connection);
    }

    memset(api_connection, 0, offsetof(api_connection_t, connected));

    uv_mutex_lock(&connections_mutex);
    LIST_INSERT_HEAD(&unconnected_head, api_connection, unconnected);
    uv_mutex_unlock(&connections_mutex);
}


void alloc_buffer(uv_handle_t *handle, size_t size, uv_buf_t *buf) {
    buf->base = calloc(1, size);
    buf->len = size;
}