#ifndef PROXYSERVER_H
#define PROXYSERVER_H

#ifdef __cpluplus
extern "C" {
#endif

#define _GNU_SOURCE

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/queue.h>

#include "uv.h"
#include "http_parser.h"
#include "uv_tls.h"

#ifndef ULLONG_MAX
# define ULLONG_MAX ((uint64_t) -1) /* 2^64-1 */
#endif

// makes CLion happy
#include "../deps/http-parser/http_parser.h"
#include "../deps/libuv/include/uv.h"
#include "../deps/evt-tls/evt_tls.h"
#include "../deps/evt-tls/sample/libuv-tls/uv_tls.h"

#define CLIENT_ID_LENGTH 26

#define RESPONSE_OK "HTTP/1.1 200 OK"
#define RESPONSE_FORBIDDEN "HTTP/1.1 403 Forbidden"
#define RESPONSE_CONNECTION_CLOSE "Content: close"
#define RESPONSE_CONNECTION_KEEP_ALIVE "Content: keep-alive"
#define RESPONSE_CONTENT_LENGTH "Content-Length: 0"
#define RESPONSE_SERVER "Server: DataDome/ProxyServer"
#define RESPONSE_X_DataDomeResponse_OK "X-DataDomeResponse: 200"
#define RESPONSE_X_DataDomeResponse_FORBIDDEN "X-DataDomeResponse: 403"
#define RESPONSE_PRAGMA "Pragma: no-cache"
#define RESPONSE_X_DataDome "X-DataDome: protected"
#define RESPONSE_X_DataDome_headers "X-DataDome-headers: X-DataDome pragma"
#define RN "\r\n"

#define MAX_HTTP_BUFFERS 64
#define MAX_HTTP_HEADERS 32
#define MAX_HTTP_BODIES  32
#define MAX_FULL_BODY    32768

typedef struct {
    const char *field;
    const char *value;
    size_t field_length;
    size_t value_length;
} http_header_t;

typedef struct {
    const char *chunk;
    size_t length;
} http_body_t;

typedef struct message_s message_t;

typedef struct {
    uv_write_t req;
    uv_stream_t stream;
    uv_timer_t timer;

    uv_buf_t response;

    int keep_alive;
    int sent_response;
    int closing;
    uint64_t content_length;
    int body_is_final;

    message_t *message;
} http_request_t;

struct message_s {
    http_parser parser;

    int last_buffer;
    uv_buf_t buffers[MAX_HTTP_BUFFERS];

    const char *url;
    size_t url_len;

    int header_lines;
    http_header_t headers[MAX_HTTP_HEADERS];

    http_body_t bodies[MAX_HTTP_BODIES];
    int last_body;

    http_request_t *request;

    char full_body[MAX_FULL_BODY];
    size_t full_body_len;

    char *new_client_id;


    LIST_ENTRY(message_s) free_messages;
    TAILQ_ENTRY(message_s) used_messages;
};

struct api_connection_s {
    struct addrinfo hints;
    uv_getaddrinfo_t resolver;

    uv_connect_t cr;
    uv_write_t req;
    uv_stream_t stream;

    uv_tls_t  sclient;

    int is_connected;

    message_t *message;

    uv_buf_t  buf;

    int send_last_buffer;

    int keep_alive;

    uint64_t content_length;
    int body_is_final;

    LIST_ENTRY(api_connection_s) connected;
    LIST_ENTRY(api_connection_s) unconnected;
};

typedef struct api_connection_s api_connection_t;

void thread_worker(void *data);

void api_connection_open(api_connection_t *api_connection);

void tcp_new_connection_cb(uv_stream_t *server, int status);

void tcp_read_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf);

void tcp_write_cb(uv_write_t *req, int status);

int http_message_begin_cb(http_parser *parser);

int http_url_cb(http_parser *parser, const char *chunk, size_t len);

int http_header_field_cb(http_parser *parser, const char *chunk, size_t len);

int http_header_value_cb(http_parser *parser, const char *chunk, size_t len);

int http_headers_complete_cb(http_parser *parser);

int http_body_cb(http_parser *parser, const char *chunk, size_t len);

void on_tcp_tls_read(uv_stream_t *stream, ssize_t nrd, const uv_buf_t *data);

void on_resolved(uv_getaddrinfo_t *resolver, int status, struct addrinfo *res);

void on_remote_proxy_tls_connection(uv_connect_t *req, int status);

void on_remote_proxy_tls_connection_read(uv_stream_t *stream, ssize_t nrd, const uv_buf_t *data);

void on_remote_tls_connection(uv_connect_t *server, int status);

void on_remote_tls_handshake(uv_tls_t *ut, int status);

void on_remote_connection(uv_connect_t *server, int status);

void send_message_cb(uv_async_t* handle);

void send_to_remote(api_connection_t *api_connection);

void tcp_write_tls_remote_cb(uv_tls_t *tls, int status);

void tcp_write_remote_cb(uv_write_t *req, int status);

void tcp_write_remote(api_connection_t *api_connection);

void tcp_read_tls_remote_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf);

void tcp_read_remote_cb(uv_stream_t *stream, ssize_t nread, const uv_buf_t *buf);

void tcp_read_remote(api_connection_t *api_connection, ssize_t nread, const uv_buf_t *buf);

int api_http_message_begin_cb(http_parser *parser);

int api_http_header_field_cb(http_parser *parser, const char *chunk, size_t len);

int api_http_header_value_cb(http_parser *parser, const char *chunk, size_t len);

int api_http_headers_complete_cb(http_parser *parser);

int api_http_body_cb(http_parser *parser, const char *chunk, size_t len);

void time_is_up(uv_timer_t* handle);

int fast_response(http_request_t *http_request);

void cleanup(http_request_t *http_request);

void reconnect_connection_timer_cb(uv_timer_t* handle);

void api_cleanup(api_connection_t *api_connection);

void api_connection_close(api_connection_t *api_connection);

void connection_closed(uv_handle_t* handle);

void api_connection_tls_closed(uv_handle_t* handle);

void api_connection_closed(uv_handle_t* handle);

void alloc_buffer(uv_handle_t *handle, size_t size, uv_buf_t *buf);

#ifdef __cpluplus
}
#endif //extern C

#endif //PROXYSERVER_H
